package edu.swe622.fers.model;

import java.io.Serializable;
import java.util.ArrayList;

/**
 * @author Naeem
 */
public class Location implements Serializable {	
	private static final long serialVersionUID = -1993069078992059692L;

	public Location(double cordX2, double cordY2) {
		this.cordX = cordX2;
		this.cordY = cordY2;
	}

	public Location() {
		
	}
	
	public Location(Location that) {
		this.cordX = that.cordX;
		this.cordY = that.cordY;
	}

	public double getCordX() {
		return cordX;
	}

	public double getCordY() {
		return cordY;
	}

	/**
	 * Computes the distance between to locations
	 * @param location1 first location
	 * @param location2 second location
	 * @return the manhattan distance between them
	 * @throws LocationMismatch the system must provide unify location types
	 */
	public double howFarIs(Location location2) throws LocationMismatch {
		if(!(location2 instanceof Location))
			throw new LocationMismatch();
		
		Location loc2 = (Location) location2;
		
		return Math.abs(cordX - loc2.cordX) + Math.abs(cordY - loc2.cordY);
	}
	
	/**
	 * Tells if source should be increased to approach the destination or vice versa. This method makes sense when source != destination.
	 * @param from source
	 * @param to destination
	 * @return 1: increase, -1: decrease, never 0: never called on equal ones
	 */
	protected int getDirection(double from, double to) {
		if(from < to)
			return 1;
		
		return -1;
	}

	/**
	 * Moves source one step (speed) toward the destination.
	 * @param destination The place where we should be 
	 * @param distancePerSecond The speed of the vehicle (not zero)
	 * @return If we have received to destination
	 * @throws InfiniteLoopException If the speedis equal to zero we will never be there
	 * @throws LocationMismatch LocationMismatch the system must provide unify location types
	 */
	public boolean moveToward(Location destination, double distancePerSecond) throws InfiniteLoopException, LocationMismatch {
		if(!(destination instanceof Location))
			throw new LocationMismatch();
		
		Location dst = (Location) destination;

		if(distancePerSecond <= 0)
			throw new InfiniteLoopException();
		
		if(equals(dst))
			return true;
		
		//How many steps we have toward the destination
		double step = distancePerSecond;		
		
		//How far the destination is
		double howFar = Math.abs(dst.cordX - cordX);
		
		if(howFar >= step) {
			cordX += step * getDirection(cordX, dst.cordX); //Going toward the destination
			return equals(dst);
		}
		
		//else
		if(howFar < step) {
			cordX = dst.cordX;
			
			step -= howFar;
			howFar = Math.abs(dst.cordY - cordY);
			
			if(howFar >= step) {
				cordY += step * getDirection(cordY, dst.cordY);
				return equals(dst);
			}
			
			//else
			if(howFar < step) {
				cordY = dst.cordY;
				return equals(dst); //equals(dst) = true
			}
		}
		
		throw new InfiniteLoopException();
	}
	
	/**
	 * Flips the direction of movement
	 * @param goToMission if it is on mission or going to return
	 */
	public synchronized void flipDestination(boolean goToMission) {
		if(forward != goToMission) {		
			forward = !forward;
			
			if(forward)
				currentDestination++;
			else
				currentDestination--;
		}
	}
	
	/**
	 * Moves source one step (speed) in the precalculated path.
	 * @return If we have received to destination
	 * @throws InfiniteLoopException If the speedis equal to zero we will never be there
	 * @throws LocationMismatch LocationMismatch the system must provide unify location types or there is no precalculated path
	 */
	public synchronized boolean followPath() throws InfiniteLoopException, LocationMismatch {
		if(currentDestination < -1 || currentDestination > thePath.size())
			throw new InfiniteLoopException();
		
		System.out.println("Direction: " + forward + ", Current Destination: " + currentDestination);
		
		if(forward) {
			if(currentDestination == thePath.size()) //We have received to final destination in previous steps
				return true;
			else {
				if(moveToward((Location)thePath.get(currentDestination), distanceIn1Second)) { //Passed the current step
					currentDestination++;
				}

				if(currentDestination == thePath.size()) //We have just received to final destination
					return true;
				else
					return false;
			}
		} else {
			if(currentDestination == -1) //We have received to final destination in previous steps
				return true;
			else {
				if(moveToward((Location)thePath.get(currentDestination), distanceIn1Second)) { //Passed the current step
					currentDestination--;
				}

				if(currentDestination == -1) //We have just received to final destination
					return true;
				else
					return false;
			}
		}		
	}

	/**
	 * When the going to destination is not straight forward user should first make a call to this function and later on call
	 * the followPath 
	 * @param path the final destination
	 * @param distancePerSecond the speed in this journey
	 */
	public void initPath(ArrayList<Location> path, double distancePerSecond) {
		thePath = path;
		distanceIn1Second = distancePerSecond;
	}

	@Override
	public Location clone(){ return new Location(cordX, cordY);}

	public boolean equals(Object other) {
		if(this == other)
			return true;
		
		if(!(other instanceof Location))
			return false;
		
		Location that = (Location)other;
		
		return ( ( Math.abs(this.cordX - that.cordX) < 0.00001) && ( Math.abs(this.cordY - that.cordY) < 0.00001) );
	}

	public int hashCode() {
		return (int)(31 * cordX + 63 * cordY);
	}

	public String toString() {
		return "Location-cordX: " + cordX + ", cordY: " + cordY;
	}
	
	protected double cordX;
	protected double cordY;
	protected ArrayList thePath = null;
	protected double distanceIn1Second = 1;
	protected int currentDestination = -1;
	protected boolean forward = false;
	/**
	 * @param cordX the cordX to set
	 */
	public void setCordX(double cordX) {
		this.cordX = cordX;
	}

	/**
	 * @param cordY the cordY to set
	 */
	public void setCordY(double cordY) {
		this.cordY = cordY;
	}
}
